"""Regras regex para atos de Reversões."""

import re
from dodfminer.extract.polished.acts.base import Atos


class Revertions(Atos):

    def __init__(self, text, backend):
        super().__init__(text, backend)

    def _regex_flags(self):
        return re.IGNORECASE

    def _act_name(self):
        return "Reversão"

    def _props_names(self):
        return ["Tipo do Ato", "SEI", "Nome", "Matricula", "Cargo", "Classe",
                "Padrao", "Quadro", "Fundamento Legal", "Orgao",
                "Vigencia", "Matricula SIAPE"]

    def _rule_for_inst(self):
        start = r"(reverter\sa\satividade[,|\s])"
        body = r"([\s\S]*?"
        end = r"(?<!lei)\s(?:[0-9|\s]*?[.|-]\s?)+?[0-9|\s]"
        end2 = r"*/\s?[0-9|\s]*-?\s?[0-9|\s]*[.|,])"
        return start + body + end + end2

    def _prop_rules(self):
        sei = r"(?<!lei)\s((?:[0-9|\s]*?[.|-]\s?)+?[0-9|\s]*/\s"
        sei2 = r"?[0-9|\s]*-?\s?[0-9|\s]*)[.|,]"
        org = r"Lotacao:|Quadro\sde\sPessoal\sd[a|e|o]([\s\S]*?)[.|,]"
        rules = {"sei": sei + sei2,
                 "nome": r"\s([^,]*?),\smatricula",
                 "matricula": r"matricula\s?n?o?\s([\s\S]*?-[\s\S]*?)[,]",
                 "cargo": r"(?:Cargo|Carreira)\sde([\s\S]*?)\,",
                 "classe": r"(?:([^,]*?)\sclasse,)?(?(1)|classe\s([\s\S]*?),)",
                 "padrao": r"[p|P]adr[a|ã]o\s([\s\S]*?),",
                 "quadro": r"d?[e|a|o]?(Quadro[\s\S]*?)[,|;|.]",
                 "fundamento": r"nos\stermos\sdo\s([\s\S]*?),\sa?\s",
                 "orgao": org,
                 "vigencia": "",
                 "siape": r"siape\sn?o?\s([\s\S]*?)[,| | .]"}

        return rules



"""Extract Title and Subtitles."""

# TODO: Improve docummentation
# TODO: Remove global variables and functions

import os
import re
import json
import fitz
import operator

from typing import List
from functools import reduce
from collections import namedtuple

import dodfminer.extract.pure.utils.title_filter as title_filter

Box = namedtuple("Box", "x0 y0 x1 y1")
BBox = namedtuple("BBox", "bbox")
TitlesSubtitles = namedtuple("TitlesSubtitles", "titles subtitles")
TextTypeBboxPageTuple = namedtuple(
    "TextTypeBboxPageTuple", "text type bbox page")
_DODF_WIDTH = 765
_DODF_HEIGHT = 907

_TRASH_WORDS = [
    "SUMÁRIO",
    "DIÁRIO OFICIAL",
    "SEÇÃO (I|II|III)",
]

_TRASH_COMPILED = re.compile('|'.join(_TRASH_WORDS))

_TYPE_TITLE, _TYPE_SUBTITLE = "title", "subtitle"
_TITLE_MULTILINE_THRESHOLD = 10


def load_blocks_list(path):
    """Loads list of blocks list from the file specified.

    Args:
        path: string with path to DODF pdf file

    Returns:
        A list with page blocks, each element being a list with its
        according page blocks.

    """
    doc = fitz.open(path)
    return [p.getTextPage().extractDICT()['blocks'] for p in doc]


def group_by_column(elements, width=_DODF_WIDTH):
    """Groups elements by its culumns.
    The sorting assumes they are on the same page
    and on a 2-column layout.

    Essentially a "groupby" where the key is the page number of each span.

    Args:
        elements: Iterable[TextTypeBboxPageTuple] sorted by its page
                  number to be grouped.

    Returns:
        A dict with spans of each page, being keys the page numbers.

    """
    left_right = [[], []]
    MID_W = width / 2
    for i in elements:
        if i.bbox.x0 <= MID_W:
            left_right[0].append(i)
        else:
            left_right[1].append(i)
    return left_right


def group_by_page(elements):
    """Groups elements by page number.

    Essentially a "groupby" where the key is the page number of each span.

    Args:
        elements: Iterable[TextTypeBboxPageTuple] sorted by its page
                  number to be grouped.

    Returns:
        A dict with spans of each page, being keys the page numbers.

    """
    page_elements = {}
    for page_num in set(map(lambda x: x.page, elements)):
        page_elements[page_num] = []
    for el in elements:
        page_elements[el.page].append(el)
    return page_elements


def sort_by_column(elements, width=_DODF_WIDTH):
    """Sorts list elements by columns.

    Args:
        elements: Iterable[TextTypeBboxPageTuple].
        width: the page width (the context in which all list elements
            were originally).

    Returns:
        List[TextTypeBboxPageTuple] containing the list elements
        sorted according to:
            1. columns
            2. position on column
        Assumes a 2-column page layout. All elements on the left column will
        be placed first of any element on the right one. Inside each columns,
        reading order is expected to be kept.

    """
    left_right = group_by_column(elements, width)

    # Sort by height
    ordenado = (sorted(i, key=lambda x: x.bbox.y0) for i in left_right)
    return reduce(operator.add, ordenado)


def _invert_TextTypeBboxPageTuple(textTypeBboxPageTuple):
    """Reverses the type between _TYPE_TITLE and _TYPE_SUBTITLE.

    Args:
        textTypeBboxPageTuple: instance of TextTypeBboxPageTuple.

    Returns:
        copy of textTypeBboxPageTuple with its type field reversed.

    """
    text, _type, bbox, page = textTypeBboxPageTuple
    return TextTypeBboxPageTuple(text, _TYPE_TITLE if _type is _TYPE_SUBTITLE
                                 else _TYPE_SUBTITLE, bbox, page)


def _extract_bold_upper_page(page):
    """Extracts page content which have bold font and are uppercase.

    Args:
        page: fitz.fitz.Page object to have its bold content extracted.

    Returns:
        A list containing all bold (and simultaneously upper)
        content at the page.

    """
    lis = []
    for bl in page.getTextPage().extractDICT()['blocks']:
        for line in bl['lines']:
            for span in line['spans']:
                flags = span['flags']
                txt: str = span['text']
                cond1 = flags in title_filter.BoldUpperCase.BOLD_FLAGS
                if cond1 and txt == txt.upper():
                    span['bbox'] = Box(*span['bbox'])
                    span['page'] = page.number
                    del span['color']
                    del span['flags']
                    lis.append(span)
    return lis


def _extract_bold_upper_pdf(path):
    """Extracts bold content from DODF pdf.

    Args:
        path: path to a DODF pdf file

    Returns:
        a list of list of bold span text

    """
    doc = fitz.open(path)
    return [_extract_bold_upper_page(page) for page in doc]


def sort_2column(elements):
    """Sorts TextTypeBboxPageTuple iterable.

    Sorts sequence of TextTypeBboxPageTuple objects, assuming a full 2-columns
    layout over them.

    Args:
        elements: Iterable[TextTypeBboxPageTuple]
    Returns:
        dictionary mapping page number to its elements sorted by column
        (assumig there are always 2 columns per page)
    """
    by_page = group_by_page(elements)
    ordered_by_page = {idx: sort_by_column(elements) for idx, elements in
                       sorted(by_page.items())}
    return ordered_by_page


def _get_titles_subtitles(elements):
    """Extracts titles and subtitles from list.

    Warning:
        Based on font size and heuristic.

    Args:
        titles_subtitles: a list of dict all of them having the keys:
            size -> float
            text -> str
            bbox -> Box
            page -> int

    Returns:
        TitlesSubtitles[List[TextTypeBboxPageTuple],
                        List[TextTypeBboxPageTuple]].

    """
    # mainly to remove "DISTRITO FEDERAL" trash below
    elements = sorted(
        elements, key=lambda d: d['size'], reverse=True)

    # Usually part of "Diário Oficial do DISTRITO FEDERAL"
    if "DISTRITO FEDERAL" in elements[0]['text']:
        del elements[0]

    # heuristic part: which font is the one use by titles?

    guessed_title_font_size = elements[min(
        2, len(elements) - 1)]['size']
    titles = []
    previous_element = elements[0]

    while elements:
        current_element = elements[0]
        if guessed_title_font_size == current_element['size']:
            if titles:
                cond1 = abs(
                    previous_element['bbox'].y1 - current_element['bbox'].y0) < _TITLE_MULTILINE_THRESHOLD
                cond2 = previous_element['page'] == current_element['page']

                # Titles must be algo in the same column

                column_grouped = group_by_column((BBox(previous_element['bbox']), BBox(current_element['bbox'])))
                cond3 = not (column_grouped[0] and column_grouped[1])
                if cond1 and cond2 and cond3:
                    titles[-1][0].append(current_element['text'])
                else:
                    titles.append([[current_element['text']], _TYPE_TITLE,
                                   current_element['bbox'],
                                   current_element['page']])
            else:
                titles.append([[current_element['text']], _TYPE_TITLE,
                               current_element['bbox'],
                               current_element['page']])
            elements = elements[1:]
        else:
            break
        previous_element = current_element

    # Titles with more than one line should be a single string

    titles = [TextTypeBboxPageTuple("\n".join(i[0]), *i[1:]) for i in titles]
    sub_titles = []
    # if the elements is over, there are no subtitles
    if elements:
        size = elements[0]['size']

        # PS: majority of subtitles uses only 1 line. Hard to distinguish
        while elements:
            current_element = elements[0]
            # TODO deal with cases like "DEPARTAMENTO DE ESTRADAS DE
            # RODAGEM DO DISTRITO FEDERAL" (5/1/2005)
            if size == current_element['size']:
                sub_titles.append((current_element['text'], _TYPE_SUBTITLE,
                                   current_element['bbox'],
                                   current_element['page']))
            else:   # this and next elements has others font sizes;
                break
            elements = elements[1:]
        sub_titles = [TextTypeBboxPageTuple(*i) for i in sub_titles]

    # Sometimes heuristic fails. However, the fix below seems to work on most
    # cases.
    # Happens mostly when there are only one title and other stuffs.

    if not titles and sub_titles:
        return TitlesSubtitles([_invert_TextTypeBboxPageTuple(i) for i in sub_titles], titles)
    else:
        return TitlesSubtitles(titles, sub_titles)


def _get_titles_subtitles_smart(path):
    """Extracts titles and subtitles. Makes use of heuristics.

    Wraps _get_titles_subtitles, removing most of impurity
    (spans not which aren't titles/subtutles).

    Args:
        path: str with the path do DODF PDF file

    Returns:
        TitlesSubtitles(List[TextTypeBboxPageTuple],
                        List[TextTypeBboxPageTuple]).
    """
    bold_spans = reduce(operator.add, _extract_bold_upper_pdf(path))
    filtered1 = filter(title_filter.BoldUpperCase.dict_text, bold_spans)
    filtered2 = filter(lambda s: not re.search(_TRASH_COMPILED, s['text']),
                       filtered1)
    # 'calibri' as font apears sometimes, however never in titles or subtitles

    filtered3 = filter(lambda x: 'calibri' not in x['font'].lower(), filtered2)

    # TODO: check for necessity of this sorting
    ordered1 = sorted(filtered3,
                      key=lambda x: (-x['page'], x['size']),
                      reverse=True)
    return _get_titles_subtitles(ordered1)


def extract_titles_subtitles(path):
    """Extracts titles and subtitles from DODF pdf.

    Args:
        path: str indicating the path for the pdf to have its
            content extracted.

    Returns:
        List[TextTypeBboxPageTuple] containing all titles ans subtitles.

    """
    titles_subtitles = _get_titles_subtitles_smart(path)
    by_page = sort_2column(reduce(operator.add, titles_subtitles))
    return reduce(operator.add, by_page.values())


# TODO: use tuples instead of lists for ensure
# immutability and avoid unexpected behavior
# (e.g, user modifying internal state of an ExtractorTitleSubtitle
# instance through appending elements to its internals lists)
class ExtractorTitleSubtitle(object):
    """Use this class like that:
    >> path = "path_to_pdf"
    >> extractor = ExtractorTitleSubtitle(path)
    >> # To extract titles
    >> titles = extractor.titles
    >> # To extract subtitles
    >> titles = extractor.subtitles
    >> # To dump titles and subtitles on a json file
    >> json_path = "valid_file_name"
    >> extractor.dump_json(json_path)
    ."""

    _TITLE_MULTILINE_THRESHOLD = 10

    def __init__(self, path):
        """.

        Args:
            path: str indicating the path for the pdf to have its
                content extracted
        """
        self._titles_subtitles = TitlesSubtitles([], [])
        self._titles = []
        self._subtitles = []
        self._path = path
        self._cached = False
        self._json = {}
        self._hierarchy = []

    def _mount_json(self):
        """Mounts json containing titles with its associated subtitles
        and store it at self._json.

        Returns:
            self._json
        """
        i = 0
        _json = {}
        titles_subtitles = self._titles_subtitles
        limit = len(titles_subtitles)
        while i < limit:
            current_element = titles_subtitles[i]
            if current_element.type == _TYPE_TITLE:
                title = current_element.text
                _json[title] = _json.get(title, [])
                i += 1
                while i < limit:
                    current_element = titles_subtitles[i]
                    if current_element.type == _TYPE_SUBTITLE:
                        _json[title].append(current_element.text)
                        i += 1
                    else:
                        break
            else:
                raise ValueError("Does not begin with a title")
        _json = {k: tuple(val) for k, val in _json.items()}
        self._json = _json
        return self._json

    def _mount_hierarchy(self):
        """Mounts list containing titles with its associated subtitles
        and store it at self._hierarchy.

        Returns:
            self._hierarchy
        """
        i = 0
        hierarchy = []
        titles_subtitles = self._titles_subtitles
        limit = len(titles_subtitles)
        while i < limit:
            current_element = titles_subtitles[i]
            if current_element.type == _TYPE_TITLE:
                title = current_element.text
                hierarchy.append([title, []])
                i += 1
                while i < limit:
                    current_element = titles_subtitles[i]
                    if current_element.type == _TYPE_SUBTITLE:
                        hierarchy[-1][1].append(current_element.text)
                        i += 1
                    else:
                        break
            else:
                raise ValueError("Não começa com títulos")
        self._hierarchy = [TitlesSubtitles(*i) for i in hierarchy]
        return self._hierarchy

    def _do_cache(self):
        """Computes some internal attributes.

        So that no more computations are needed for them.
        Implicit called after first access to some property.
        Computes and caches the following internal attributes:
            - _titles_subtitles
            - _titles
            - _subtitles
        """
        self._titles_subtitles = tuple(extract_titles_subtitles(self._path))
        self._titles = tuple(filter(lambda x: x.type == _TYPE_TITLE,
                                    self._titles_subtitles))
        self._subtitles = tuple(filter(lambda x: x.type == _TYPE_SUBTITLE,
                                       self._titles_subtitles))
        self._cached = True

    @property
    def titles(self):
        """All titles extracted from the file speficied by self._path.

        Returns:
            List[TextTypeBboxPageTuple] each of which having its type attribute
            equals _TYPE_TITLE
        """
        if not self._cached:
            self._do_cache()
        return list(self._titles)

    @property
    def subtitles(self):
        """All subtitles extracted from the file speficied by self._path.

        Returns:
            List[TextTypeBboxPageTuple] each of which having its type attribute
            equals _TYPE_SUBTITLE
        """
        if not self._cached:
            self._do_cache()
        return list(self._subtitles)

    @property
    def titles_subtitles(self):
        """A list with titles and subtitles, sorted according to its reading order.
        """
        if not self._cached:
            self._do_cache()
        return list(self._titles_subtitles)

    @property
    def json(self):
        """All titles with its subtitles associated.

        All subtitles under the same title are at the same level.
        Deprecated. Better use `titles_subtitles` or
        `titles_subtitles_hierarchy`.
        """
        if not self._json:
            if not self._cached:
                self._do_cache()
            self._mount_json()
        # return self._json.copy()
        return {k: list(val) for k, val in self._json.items()}

    @property
    def titles_subtitles_hierarchy(self) -> TitlesSubtitles(str, List[str]):
        """All titles and subtitles extracted from the file specified by
        self._path, hierarchically organized.

        Returns:
            List[TitlesSubtitles(str, List[str])]: the titles and its
            respectively subtitles
        """
        if not self._hierarchy:
            if not self._cached:
                self._do_cache()
            self._mount_hierarchy()
        return self._hierarchy.copy()

    def dump_json(self, path):
        """Writes on file specified by path the JSON representation of titles
        and subtitles extracted.

        Dumps the titles and subtitles according to the hierarchy verified
        on the document.

        The outputfile should be specified and will be suffixed with the
        ".json" if it's not.

        Args:
            path: string containing path to .json file where the dump will
            be done. Its suffixed with ".json" if it's not.

        """
        json.dump(self.json,
                  open("{}{}".format(
                      path, (not path.endswith(".json")) * ".json"), 'w'),
                  ensure_ascii=False, indent='  ')

    def reset(self):
        """Sets cache to False and reset others internal attributes.
        Use when for some reason the internal state was
        somehow modified by user.
        """
        self._json = {}
        self._hierarchy = []
        self._cache = False


def gen_title_base(dir_path=".", base_name="titles", indent=4, forced=False):
    """Generates titles base from all PDFs immediately under dir_path directory.
    The base is generated under dir_path directory.
    Args:
        dir_path: path so base_name will contain all titles
            from PDFs under dir_path
        base_name: titles' base file name
        indent: how many spaces used will be used for indent
    Returns:
        dict containing "titles" as key and a list of titles,
            the same stored at base_name[.json]
    """
    base_name = "{}/{}".format(
        dir_path, base_name + (not base_name.endswith(".json")) * ".json")
    if os.path.exists(base_name) and not forced:
        print("Error: {} already exists".format(base_name))
        return None
    elif os.path.isdir(base_name):
        print("Error: {} ir a directory".format(base_name))
        return None

    titles = set()
    for file in filter(lambda x: not os.path.isdir(x) and x.endswith('.pdf'), os.listdir(dir_path)):
        et = ExtractorTitleSubtitle(file)
        titles_text = map(lambda x: x.text, et.titles)
        titles.update(titles_text)
    js = {"titles": list(titles)}
    json.dump(js, open("{}".format(base_name), 'w'),
              ensure_ascii=False, indent=indent*' ')

    return js


def gen_hierarchy_base(dir_path=".",
                       folder="hierarchy", indent=4, forced=False):
    """Generates json base from all PDFs immediately under dir_path directory.
    The hiearchy files are generated under dir_path directory.
    Args:
        dir_path: path so folder containing PDFs
        base_name: titles' base file name
        forced: proceed even if folder `base_name` already exists
        indent: how many spaces used will be used for indent
    Returns:
        List[Dict[str, List[Dict[str, List[Dict[str, str]]]]]]
        e.g:
        [
           { "22012019": [
                {
                  "PODER EXECUTIVO": []
                },
                {
                    "SECRETARIA DE ESTADO DE FAZENDA,\nPLANEJAMENTO, ORÇAMENTO E GESTÃO": [
                        {
                            "SUBSECRETARIA DA RECEITA": ""
                        }
                    ]
                }
            }
        ]
        In case of error trying to create `base_name` folder,
        returns None.
    """
    folder = "{}/{}".format(dir_path, folder)
    if not dir_path:
        dir_path = "."
    try:
        os.makedirs(folder, exist_ok=forced)
    except Exception as error:
        print(error)
        return None

    hierarchies = []
    for file in filter(lambda x: x.endswith('.pdf'), os.listdir(dir_path)):
        et = ExtractorTitleSubtitle("{}/{}".format(dir_path, file))
        hierarchy = et.titles_subtitles_hierarchy
        hierarchy = [
            ({d[0]: [dict([(i, '')]) for i in d[1]]})
            for d in hierarchy
        ]
        hierarchy = {file.rstrip('.pdf'): hierarchy}
        hierarchies.append(hierarchy)

        json.dump(hierarchy,
                  open("{}/{}.json".format(
                                    folder, file.rstrip('.pdf')), 'w'),
                  ensure_ascii=False, indent=indent*' ')

    return hierarchies



import re
import os
import joblib
import pandas as pd
from typing import List, Match

from dodfminer.extract.polished.acts.base import Atos


def case_insensitive(s: str):
    """Returns regular expression similar to `s` but case careless.

    Note: strings containing characters set, as `[ab]` will be transformed to `[[Aa][Bb]]`.
        `s` is espected to NOT contain situations like that.
    Args:
        s: the stringregular expression string to be transformed into case careless
    Returns:
        the new case-insensitive string
    """

    return ''.join([c if not c.isalpha() else '[{}{}]'.format(c.upper(), c.lower()) for c in s])


def remove_crossed_words(s: str):
    """Any hyfen followed by 1+ spaces are removed.
    """
    return re.sub(r'-\s+', '', s)


LOWER_LETTER = r"[áàâäéèẽëíìîïóòôöúùûüça-z]"
UPPER_LETTER = r"[ÁÀÂÄÉÈẼËÍÌÎÏÓÒÔÖÚÙÛÜÇA-Z]"

DODF = r"(DODF|[Dd]i.rio\s+[Oo]ficial\s+[Dd]o\s+[Dd]istrito\s+[Ff]ederal)"

SIAPE = r"{}\s*(?:n?.?)\s*(?P<siape>[-\d.Xx/\s]+)".format(case_insensitive("siape"))
# SIAPE = r"(?i:siape)\s*(?:n?.?)\s*(?P<siape>[-\d.Xx/\s]+)"

MATRICULA = r"(?:matr.cul.|matr?[.]?\B)[^\d]+(?P<matricula>[-\d.XxZzYz/\s]+)"
# MATRICULA = r"(?i:matr.cul.|matr?[.]?\B)[^\d]+(?P<matricula>[-\d.XxZzYz/\s]+)"
MATRICULA_GENERICO = r"(?<![^\s])(?P<matricula>([-\d.XxZzYz/\s]{1,})[.-][\dXxYy][^\d])"
MATRICULA_ENTRE_VIRGULAS = r"(?<=[A-ZÀ-Ž]{3})\s*,\s+(?P<matricula>[-\d.XxZzYz/\s]{3,}?),"

SERVIDOR_NOME_COMPLETO = r"(servidor.?|empregad.)[^A-ZÀ-Ž]{0,40}(?P<name>[A-ZÀ-Ž][.'A-ZÀ-Ž\s]{6,}(?=[,]))"
# SERVIDOR_NOME_COMPLETO = r"(?i:servidor.?|empregad.)[^A-ZÀ-Ž]{0,40}(?P<name>[A-ZÀ-Ž][.'A-ZÀ-Ž\s]{6,}(?=[,]))"
NOME_COMPLETO = r"(?P<name>['A-ZÀ-Ž][.'A-ZÀ-Ž\s]{6,}(?=[,.:;]))"

PROCESSO_NUM = r"(?P<processo>[-0-9/.]+)"
INTERESSADO = r"{}:\s*{}".format(case_insensitive("interessad."), NOME_COMPLETO)
# INTERESSADO = r"(?i:interessad.):\s*{}".format(NOME_COMPLETO)
# INTERESSADO = r"(?i:interessad.):\s*" + NOME_COMPLETO

ONUS = r"(?P<onus>\b[oôOÔ]{}\b[^.]+[.])".format(case_insensitive("nus"))
# ONUS = r"(?P<onus>\b[oôOÔ](?i:(nus))\b[^.]+[.])"


class Cessoes(Atos):
    _special_acts = ['matricula', 'cargo']

    def __init__(self, file, backend, debug=False, extra_search=True):
        self._debug = debug
        self._extra_search = extra_search
        self._processed_text = remove_crossed_words(open(file).read())
        self._raw_matches = []
        super().__init__(file, backend)

    def _act_name(self):
        return "Cessoes"

    def _load_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/models/cessoes_ner.pkl'
        return joblib.load(f_path)

    def _props_names(self):
        return ["tipo"] + list(self._prop_rules())

    def _rule_for_inst(self):
        return (
            r"([Pp][Rr][Oo][Cc][Ee][Ss][Ss][Oo][^0-9/]{0,12})([^\n]+?\n){0,2}?"
            + r"[^\n]*?[Aa]\s*[Ss]\s*[Ss]\s*[Uu]\s*[Nn]\s*[Tt]\s*[Oo]\s*:?\s*\bCESS.O\b"
            + r"([^\n]*\n){0,}?[^\n]*?(?=(?P<look_ahead>PROCESSO|Processo:|PUBLICAR|pertinentes[.]|autoridade cedente|"
            + case_insensitive('publique-se') + "))"
            # + r'(?i:publique-se)' + "))"
        )

    def _prop_rules(self):
        return {
            'interessado': INTERESSADO,
            'nome': SERVIDOR_NOME_COMPLETO,
            'matricula': MATRICULA,
            'processo': r"[^0-9]+?{}".format(PROCESSO_NUM),
            # 'processo': r"[^0-9]+?" + PROCESSO_NUM,
            'onus': ONUS,
            'siape': SIAPE,
            'cargo': r",(?P<cargo>[^,]+)",
        }

    def _find_instances(self) -> List[Match]:
        """Returns list of re.Match objects found on `self._text_no_crosswords`.

        Return:
            a list with all re.Match objects resulted from searching for
        """

        self._raw_matches = list(
            re.finditer(self._inst_rule, self._processed_text, flags=self._flags)
        )
        l = [i.group() for i in self._raw_matches]
        if self._debug:
            print("DEBUG:", len(l), 'matches')
        return l

    def _get_special_acts(self, lis_matches):
        for i, match in enumerate(self._raw_matches):
            act = match.group()
            matricula = re.search(MATRICULA, act) or \
                    re.search(MATRICULA_GENERICO, act) or \
                    re.search(MATRICULA_ENTRE_VIRGULAS, act)

            nome = re.search(self._rules['nome'], act)
            if matricula and nome:
                offset = matricula.end()-1 if 0 <= (matricula.start() - nome.end()) <= 5 \
                            else nome.end() - 1
                cargo, = self._find_props(r",(?P<cargo>[^,]+)", act[offset:])
            else:
                cargo = "nan"

            lis_matches[i]['matricula'] = matricula.group('matricula') if matricula \
                                        else "nan"
            lis_matches[i]['cargo'] = cargo


    def _find_props(self, rule, act):
        """Returns named group, or the whole match if no named groups
                are present on the match.
        Args:
            match: a re.Match object
        Returns: content of the unique named group found at match,
            the whole match if there are no groups at all or raise
            an exception if there are more than two groups.
        """
        match = re.search(rule, act, flags=self._flags)

        if match:
            keys = list(match.groupdict().keys())
            if len(keys) == 0:
                return match.group()
            elif len(keys) > 1:
                raise ValueError("Named regex must have AT MOST ONE NAMED GROUP.")
            if self._debug:
                print('key: ', keys[0])
            return match.group(keys[0]),
        else:
            return "nan"


    def _acts_props(self):
        acts = []
        for raw in self._raw_acts:
            act = self._act_props(raw)
            acts.append(act)
        if self._extra_search:
            self._get_special_acts(acts)
        return acts


    def _extract_instances(self) -> List[Match]:
        found = self._find_instances()
        self._acts_str = found.copy()
        return found



# coding=utf-8

"""Extract content from DODFS and export to JSON.

Contains class ContentExtractor which have to public functions
avaiable to extract the DODF to JSON

Usage example::

    pdf_text = ContentExtractor.extract_text(file)
    ContentExtractor.extract_to_txt(folder)

"""

import os
import json
import unicodedata

from pathlib import Path

import fitz

from dodfminer.extract.pure.utils.title_extractor import ExtractorTitleSubtitle
from dodfminer.extract.pure.utils.box_extractor import get_doc_text_boxes

RESULTS_PATH = "results/"
RESULTS_PATH_JSON = "results/json"
RESULTS_PATH_TXT = "results/txt"


class ContentExtractor:
    """Extract content from DODFs and export to JSON.

    Extracts content from DODF using as suport the title and subtitle
    databases (which runs using MuPDF) and the Tesseract OCR library. All the
    contents are exported to a JSON in which it's keys are DODFs titles or
    subtitles, and its values are the correspondent content.

    Note:
        This class is not constructable, it cannot generate objects.

    """

    @classmethod
    def extract_text(cls, file, single=False, block=False, sep=' ', norm='NFKD'):
        """Extract block of text from file

        Args:
            file: The DODF to extract the titles.
            single: output content in a single file in the file directory
            block: Extract the text as a list of text blocks
            sep: The separator character between each block of text
            norm: Type of normalization applied to the text

        Returns:
            A list of arrays if block is True or strign otherwise.
            Each array in the list have 5 values: the first four
            are the coordinates of the box (x0, y0, x1, y1) end the last
            is the text from the box.

            The string returned correspond to the text from all PDF.

        """
        drawboxes_text = ''
        list_of_boxes = []
        pymu_file = fitz.open(file)
        for textboxes in get_doc_text_boxes(pymu_file):
            for text in textboxes:
                if int(text[1]) != 55 and int(text[1]) != 881:
                    if block:
                        norm_text = cls._normalize_text(text[4], norm)
                        list_of_boxes.append((text[0], text[1], text[2],
                                              text[3], norm_text))
                    else:
                        drawboxes_text += (text[4] + sep)

        if block:
            return list_of_boxes if not single else cls._save_single_file(file, 'json', json.dumps(list_of_boxes))

        drawboxes_text = cls._normalize_text(drawboxes_text, norm)
        return drawboxes_text if not single else cls._save_single_file(file, 'txt', drawboxes_text)

    @classmethod
    def extract_structure(cls, file, single=False, norm='NFKD'):
        """Extract boxes of text with your respective titles.

        Args:
            file: The DODF to extract the titles.
            single: output content in a single file in the file directory
            norm: Type of normalization applied to the text

        Returns:
            A dictionaty with the blocks organized by title

            Example::

                {
                    "Title": [
                        [
                            x0,
                            y0,
                            x1,
                            y1,
                            "Text"
                        ]
                    ],
                    ...
                }

        """
        content_dict = {}
        try:
            title_base = cls._extract_titles(file).json.keys()
        except Exception as e:
            cls._log(e)
        else:
            boxes = cls.extract_text(file, block=True, norm=norm)
            first_title = False
            is_title = False
            actual_title = ''
            for box in boxes:
                text = box[4]
                for title in title_base:
                    title.replace("\n", "")
                    if text == title:
                        first_title = True
                        is_title = True
                        actual_title = title
                        if title not in content_dict.keys():
                            content_dict.update({title: []})
                    else:
                        is_title = False

                if first_title and not is_title:
                    if int(box[1]) != 55 and int(box[1]) != 881:
                        content_dict[actual_title].append(box[:5])

            return content_dict if not single else cls._save_single_file(file, 'json', json.dumps(content_dict))
    @classmethod
    def extract_to_txt(cls, folder='./', norm='NFKD'):
        """Extract information from DODF to TXT.

        For each pdf file in data/dodfs, extract information from the
        pdf and output it to txt.

        """
        pdfs_path_list = cls._get_pdfs_list(folder)
        cls._create_single_folder(os.path.join(folder, RESULTS_PATH))
        cls._create_single_folder(os.path.join(folder, RESULTS_PATH_TXT))
        txt_path_list = cls._get_txt_list(folder)

        for file in pdfs_path_list:
            pdf_name = os.path.splitext(os.path.basename(file))[0]
            if pdf_name not in txt_path_list:
                cls._log(pdf_name)
                text = cls.extract_text(file, norm=norm)
                t_path = cls._struct_subfolders(file, False, folder)
                f = open(t_path, "w")
                f.write(text)
            else:
                cls._log("TXT already exists")

    @classmethod
    def extract_to_json(cls, folder='./',
                        titles_with_boxes=False, norm='NFKD'):
        """Extract information from DODF to JSON.

        Args:
            folder: The folder containing the PDFs to be extracted
            titles_with_boxes: Extract with titles
            norm: What normalization to use. Normalizations can be found
                  unicodedata library

        For each pdf file in data/dodfs, extract information from the
        pdf and output it to json.

        """
        # Get list of all downloaded pdfs
        pdfs_path_list = cls._get_pdfs_list(folder)
        # Get list of existing json to not repeat work
        json_path_list = cls._get_json_list(folder)

        cls._create_single_folder(os.path.join(folder, RESULTS_PATH))
        cls._create_single_folder(os.path.join(folder, RESULTS_PATH_JSON))

        for file in pdfs_path_list:
            pdf_name = os.path.splitext(os.path.basename(file))[0]
            # We do not want the system to repeat itself doing the same work
            if pdf_name not in json_path_list:
                # low cost extractions
                if os.path.getsize(file) < 30000000:  # Remove in future.
                    # Remove images that might still there from previous exec
                    cls._log(pdf_name)
                    if titles_with_boxes:
                        content = cls.extract_structure(file, norm=norm)
                    else:
                        content = cls.extract_text(file, block=True, norm=norm)
                    j_path = cls._struct_subfolders(file, True, folder)
                    json.dump(content, open(j_path, "w", encoding="utf-8"),
                              ensure_ascii=False)
            else:
                cls._log("JSON already exists")

    @classmethod
    def _save_single_file(cls, file_path, type, content):
        open(file_path.replace('pdf', type), 'w+').write(content)

    @classmethod
    def _normalize_text(cls, text, form='NFKD'):
        """Normalize text.

        Args:
            text: The text to be normalized.
            form: The normalized form accordingly to the unicodedata library'.

        Returns:
            A string with the normalized text.

        """
        normalized = unicodedata.normalize(form, text).encode('ascii', 'ignore').decode('utf8')
        return normalized

    @classmethod
    def _extract_titles(cls, file):
        """Extract titles and subtitles from the DODF.

        Args:
            file: The DODF to extract the titles.

        Returns:
            An object of type ExtractorTitleSubtitle, in which have the
            attributes:

            titles: get all titles from pdf
            subtitle: get all subtitles from pdf

        Raises:
            Exception: error in extracting titles from pdf

        """
        try:
            title_database = ExtractorTitleSubtitle(file)
            cls._log(file)
        except Exception as e:
            cls._log(f"Error in extracting files from {file}: {e}")
            raise
        else:
            return title_database

    @classmethod
    def _get_pdfs_list(cls, folder):
        """Get DODFs list from the path.

        Returns:
            A list of DODFS' pdfs paths.

        """
        pdfs_path_list = []
        for dp, _, fn in os.walk(os.path.expanduser(os.path.join(folder))):
            for f in fn:
                if '.pdf' in f:
                    pdfs_path_list.append(os.path.join(dp, f))

        return pdfs_path_list

    @classmethod
    def _get_json_list(cls, folder):
        """Get list of exisiting JSON from the path.

        Returns:
            A list of all exisiting JSON's.

        """
        aux = []
        for dp, _, fn in os.walk(os.path.expanduser(os.path.join(folder, RESULTS_PATH_JSON))):
            for f in fn:
                aux.append(os.path.join(dp, f))

        json_path_list = []
        for file in aux:
            json_path_list.append(os.path.splitext(os.path.basename(file))[0])

        return json_path_list

    @classmethod
    def _get_txt_list(cls, folder):
        """Get list of exisiting TXT from the path.

        Returns:
            A list of all exisiting TXT's.

        """
        aux = []
        for dp, _, fn in os.walk(os.path.expanduser(os.path.join(folder, RESULTS_PATH_TXT))):
            for f in fn:
                aux.append(os.path.join(dp, f))

        txt_path_list = []
        for file in aux:
            txt_path_list.append(os.path.splitext(os.path.basename(file))[0])

        return txt_path_list

    @classmethod
    def _struct_subfolders(cls, path, json_f, folder):
        """Create directory for the JSON.

        Using the same standart from the DODFs directory tree. Create the
        JSONs folder as same.

        Raises:
            FileExistsError: The folder being created is already there.

        Return:
            The path created for the JSON to be saved.

        """
        type_f = '.json' if json_f else '.txt'
        res_path = RESULTS_PATH_JSON if json_f else RESULTS_PATH_TXT

        path = path.replace(folder, "")
        splited = path.split('/')
        basename = splited[-1].split('.')
        basename = basename[0] + type_f
        splited[-1] = basename
        final_path = '/'.join(splited[1:])
        final_path = os.path.join(folder, res_path, final_path)
        path = Path(os.path.join(folder, res_path, *splited[:-1]))
        try:
            path.mkdir(parents=True)
        except FileExistsError:
            pass

        return final_path

    @classmethod
    def _create_single_folder(cls, path):
        """Create a single folder given the directory path.

        This function might create a folder, observe that the folder already
        exists, or raise an error if the folder cannot be created.

        Args:
            path: The path to be created

        Raises:
            OSError: Error creating the directory.

        """
        if os.path.exists(path):
            cls._log(os.path.basename(path) + " folder already exist")
        else:
            try:
                os.mkdir(path)
            except OSError as error:
                cls._log("Exception during the directory creation")
                cls._log(str(error))
            else:
                basename = os.path.basename(path)
                cls._log(basename + " directory successful created")

    @classmethod
    def _log(cls, msg):
        print(f"[EXTRACTOR] {msg}")



"""Regras regex para ato de Substituição."""

import re
from dodfminer.extract.polished.acts.base import Atos


class Substituicao(Atos):

    def __init__(self, file, backend):
        super().__init__(file, backend)

    # def _load_model(self):
    #     f_path = os.path.dirname(__file__)
    #     f_path += '/models/substituicao_ner.pkl'
    #     return joblib.load(f_path)

    def _regex_flags(self):
        return re.IGNORECASE

    def _act_name(self):
        return "Substituição de Funções"

    def _props_names(self):
        return ["Tipo do Ato", "Nome do Servidor Substituto",
                "Matrícula do Servidor Substituto",
                "Nome do Servidor a ser Substituido",
                "Matrícula do Servidor a ser Substituido"
                "Cargo", "Símbolo do cargo do servidor substituto",
                "Cargo comissionado objeto da substituição",
                "Símbolo do cargo do objeto da substituição",
                "Símbolo do cargo comissionado objeto da substituição",
                "Hierarquia da Lotação", "Órgão", "Data Inicial da Vigência",
                "Data Final de Vigência", "Matrícula SIAPE", "Motivo"]

    def _rule_for_inst(self):
        start = r"(DESIGNAR)"
        body = r"([\s\S]*?)"
        end = r"\.\s"
        return start + body + end

    def _prop_rules(self):

        rules = {"Nome Serv Substituto": r"(^[A-ZÀ-Ž\s]+[A-ZÀ-Ž])",
                 "Matricula Serv Substituto": r"(?:^[A-ZÀ-Ž\s]+[A-ZÀ-Ž])\s[\s\S]*?\smatr[í|i]cula\s?n?o?\s([\s\S]*?)[,|\s]",
                 "Nome do Servidor a ser Substituido": r"para\ssubstituir\s([A-ZÀ-Ž\s]+[A-ZÀ-Ž])",
                 "Matrícula do Servidor a ser Substituido": r"matr[í|i]cula\s?n?o?\s(?:[\s\S]*?)[\s\S]*?matr[í|i]cula\s?n?o?\s([\s\S]*?),",
                 "cargo": r"para\ssubstituir\s(?:[A-ZÀ-Ž\s]+[A-ZÀ-Ž]),\smatr[í|i]cula\s?n?o?\s(?:[\s\S]*?),\s([\s\S]*?),",
                 "Símbolo do cargo do servidor substituto": r"[S|s][í|i]mbolo\s?n?o?\s([\s\S]*?)[,|\s]",
                 "Cargo comissionado objeto da substituição": "",
                 "Símbolo do cargo do objeto da substituição": r"[S|s][í|i]mbolo\s?n?o?\s(?:[\s\S]*?)[\s\S]*?[S|s][í|i]mbolo\s?n?o?\s([\s\S]*?)",
                 "Hierarquia da Lotação": "",
                 "orgao": r"(?:[S|s][í|i]mbolo\s?n?o?\s(?:[\s\S]*?)[,|\s])\sde(?:[\s\S]*?),\sd[a|e|o]\s(?:[\s\S]*,?),\sd[a|e|o]\s([\s\S]*?)$",
                 "Data Inicial da Vigência": "",
                 "Data Final de Vigência": "",
                 "siape": r"[S|s][I|i][A|a][P|p][E|e]\s[N|n]?[o|O]?\s([\s\S]*?)[,| | .]",
                 "Motivo": ""}
        return rules



import re
from typing import List, Match
import pandas as pd
from dodfminer.extract.polished.acts.base import Atos


DODF = r"(DODF|[Dd]i.rio\s+[Oo]ficial\s+[Dd]o\s+[Dd]istrito\s+[Ff]ederal)"
_EDICAO_DODF = r"(\b(?i:suplement(o|ar)|extra|.ntegra))\b."
TIPO_EDICAO = r"\b(?P<tipo>(?i:extra(\sespecial)?|suplement(ar|o)))\b"
DODF_TIPO_EDICAO = DODF + r"(?P<tipo_edicao>.{0,50}?)" + _EDICAO_DODF

MONTHS = (
    r'(?i:janeiro|fevereiro|mar.o|abril|maio|junho|'
    r'julho|agosto|setembro|outubro|novembro|dezembro)'
)

FLEX_DATE = r"(?P<date>\d+\s+(?:de\s*)?" + MONTHS + \
    r"\s*(?:de\s*)?\d+|\d+[.]\d+[.]\d+|\d+[/]\d+[/]\d+)"

DODF_NUM = r"(?i:DODF|[Dd]i.rio [Oo]ficial [Dd]o [Dd]istrito [Ff]ederal)\s*[\w\W]{0,3}?(?i:n?(.mero|[.roº]{1,4})?[^\d]+?)(?P<num>\d+)"

DODF_DATE = DODF + r"[^\n\n]{0,50}?(de\s?)?" + FLEX_DATE

SIAPE = r"(?i:siape)\s*(?i:n?.?)\s*[-\d.Xx/\s]"

MATRICULA = r"(?i:matr.cul.|matr?[.]?\B)[^\d]+(?P<matricula>[-\d.XxZzYz/\s]+)"

MATRICULA_GENERICO = r"(?<![^\s])(?P<matricula>([-\d.XxZzYz/\s]{1,})[.-][\dXxYy][^\d])"

MATRICULA_ENTRE_VIRGULAS = r"(?<=[A-ZÀ-Ž]{3})\s*,\s+([-\d.XxZzYz/\s]{3,}?),"

PAGE = r"((?i:p\.|p.ginas?|p.?gs?\.?\b)(?P<page_nums>.{0,}?)(?=[,;:]|\n|\s[A-Z]|$))"

SERVIDOR_NOME_COMPLETO = r"(?i:servidora?\b.{0,40}?)(?P<name>[A-ZÀ-Ž][.'A-ZÀ-Ž\s]{7,})"

NOME_COMPLETO = r"(?P<name>[.'A-ZÀ-Ž\s]{8,})"

LOWER_LETTER = r"[áàâäéèẽëíìîïóòôöúùûüça-z]"
UPPER_LETTER = r"[ÁÀÂÄÉÈẼËÍÌÎÏÓÒÔÖÚÙÛÜÇA-Z]"

PROCESSO_MATCH = r"(?i:processo):?[^\d]{0,50}?(?P<processo>\d[-0-9./\s]*\d(?!\d))"
TIPO_DOCUMENTO = r"(?i:portaria|ordem de servi.o|instru..o)"


class SemEfeitoAposentadoria(Atos):
    _special_acts = [
        'dodf_num', 'tornado_sem_efeito_publicacao',
        'dodf_pagina', 'servidor', 'matricula',
        'cargo', 'dodf_tipo_edicao',
    ]

    _BAD_MATCH_WORDS = [
        "AVERBAR",
        "NOMEAR",
        "CONCEDER",
        "EXONERAR",
        "DESAVERBAR",
        "APOSTILAR",
        "RETIFICAR",
    ]

    def _pre_process_text(self, s):
        # Make sure words splitted accross lines are joined together
        no_split_word = s.replace('-\n', '-')
        return no_split_word.replace('\n', ' ')

    def __init__(self, file, backend, debug=False, extra_search=True,
                 nlp=None, max_length=2000):
        self._max_length = max_length
        self._debug = debug
        self._extra_search = extra_search
        self._processed_text = self._pre_process_text(open(file).read())
        self._raw_matches = []
        self._nlp = nlp
        super().__init__(file, backend)

    def _act_name(self):
        return "Atos tornados sem efeito - aposentadoria"

    def _props_names(self):
        return list(self._prop_rules())

    def _rule_for_inst(self):
        return (
            r"TORNAR SEM EFEITO"
            r"([^\n]+\n){0,10}?[^\n]*?"
            r"(tempo\sde\sservi.o|aposentadoria|aposentou|([Dd][Ee][Ss])?[Aa][Vv][Ee][Rr][Bb][Aa]..[Oo]|(des)?averb(ar?|ou))"
            r"[\d\D]{0,500}?[.]\s+"
            r"(?=[A-Z]{3})"
        )

    def _prop_rules(self):
        return {
            'tipo_documento': TIPO_DOCUMENTO,
            'processo': PROCESSO_MATCH,
            'dodf_data': DODF_DATE,
        }

    def _find_instances(self) -> List[Match]:
        """Returns list of re.Match objects found on `self._text_no_crosswords`.

        Return:
            a list with all re.Match objects resulted from searching for
        """
        head = "TORNAR SEM EFEITO"
        end = " CAFEBABE"   # so that lookeahead does not become a problem
        lis = self._processed_text.split(head)
        lis = [
            re.search(self._inst_rule, head + tex + end)
            # content before first `head` occurrence does not matter
            for tex in lis[1:]
        ]
        lis = [i for i in lis if i]

        true_positive = []
        for raw_match in lis:
            flag = True
            for bad in self._BAD_MATCH_WORDS:
                s = raw_match.group()
                if len(s) > self._max_length or bad in s:
                    flag = False
                    break
            if flag:
                true_positive.append(raw_match)

        self._raw_matches = true_positive
        if self._debug:
            print("DEBUG:", len(lis), 'generic matches')
            print("DEBUG:", len(self._raw_matches), 'true matches')
        return [i.group() for i in self._raw_matches]

    def _get_special_acts(self, lis_dict):
        for i, match in enumerate(self._raw_matches):
            act = match.group()
            curr_dict = lis_dict[i]
            dodf_date = curr_dict['dodf_data']
            dodf_num = dodf_date and re.search(DODF_NUM, dodf_date.group())
            tornado_sem_efeito_publicacao = dodf_date and \
                re.search(
                    FLEX_DATE, act[:dodf_date.start()] + act[dodf_date.end():])
            dodf_pagina = dodf_date and re.search(
                PAGE, act[dodf_date.end():][:50])

            servidor = re.search(SERVIDOR_NOME_COMPLETO, act)
            if not servidor:
                #  If it fails then a more generic regex is searched for
                dodf_mt = re.search(DODF, act)
                dodf_end = 0 if not dodf_mt else dodf_mt.end()
                servidor = re.search(NOME_COMPLETO, act[dodf_end:])
                del dodf_mt, dodf_end
                if not servidor:
                    # Appeal to spacy
                    all_cands = re.findall(NOME_COMPLETO, act)
                    cand_text = 'SEM-SERVIDOR'
                    for cand in self._nlp(', '.join([c.strip().title() for c in all_cands])).ents:
                        cand_text = cand.text

                        if cand.label_ == 'PER':
                            break
                    servidor = re.search(cand_text.upper(), act)
                    del all_cands, cand_text, cand
            end_employee = servidor.end() if servidor else 0
            matricula = re.search(MATRICULA, act[end_employee:]) or \
                re.search(MATRICULA_GENERICO, act[end_employee:]) or \
                re.search(MATRICULA_ENTRE_VIRGULAS, act[end_employee:])

            del end_employee
            if not matricula or not servidor:
                cargo = None
            else:
                servidor_start = act[servidor.start():].find(
                    servidor.group()) + servidor.start()
                matricula_start = act[matricula.start():].find(
                    matricula.group()) + matricula.start()

                # NOTE: -1 is important in case `matricula` end with `,`
                if 0 <= (matricula_start - (servidor_start + len(servidor.group()))) <= 5:
                    # cargo does not fit between 'servidor' and 'matricula'
                    cargo = re.search(
                        r",(?P<cargo>[^,]+)", act[matricula_start + len(matricula.group())-1:])

                else:
                    # cargo right after employee's name

                    cargo = re.search(
                        r",(?P<cargo>[^,]+)", act[servidor_start + len(servidor.group())-1:])
                del matricula_start, servidor_start
            edicao = re.search(DODF_TIPO_EDICAO, act)
            dodf_tipo_edicao = re.search(TIPO_EDICAO, act[edicao.start()-1:edicao.end()+1])\
                if edicao else re.search("normal", "normal")
            curr_dict['dodf_num'] = dodf_num
            curr_dict['tornado_sem_efeito_publicacao'] = tornado_sem_efeito_publicacao
            curr_dict['dodf_pagina'] = dodf_pagina
            curr_dict['servidor'] = servidor
            curr_dict['matricula'] = matricula
            curr_dict['cargo'] = cargo

            curr_dict['dodf_tipo_edicao'] = dodf_tipo_edicao

    def _find_props(self, rule, act):
        """Returns named group, or the whole match if no named groups
                are present on the match.
        Args:
            match: a re.Match object
        Returns: content of the unique named group found at match,
            the whole match if there are no groups at all or raise
            an exception if there are more than two groups.
        """
        match = re.search(rule, act, flags=self._flags)
        return match,

    def _group_solver(self, match):
        """Returns named group, or the whole match if no named groups
                are present on the match.
        Args:
            match: a re.Match object
        Returns: content of the unique named group found at match,
            the whole match if there are no groups at all or raise
            an exception if there are more than two groups.
        """
        if not match or isinstance(match, str):
            return "nan"
        elif match.groupdict():
            key = list(match.groupdict())[0]
            return match.group(key)
        else:
            return match.group()

    def _acts_props(self):
        acts = []
        for raw in self._raw_acts:
            act = self._act_props(raw)
            acts.append(act)
        if self._extra_search:
            self._get_special_acts(acts)
        return acts

    def _extract_instances(self) -> List[Match]:
        found = self._find_instances()
        self._acts_str = found.copy()
        return found

    def _build_dataframe(self):
        _ = re.search(self._name, self._name)
        for dic in self._acts:
            dic["tipo_ato"] = _
        data = [{k: self._group_solver(v) for k, v in act.items()}
                for act in self._acts]
        return pd.DataFrame(data)



"""Regras regex para ato de Abono de Permanencia."""

import re
import os
import joblib
from dodfminer.extract.polished.acts.base import Atos


class AbonoPermanencia(Atos):

    def __init__(self, file, backend):
        super().__init__(file, backend)

    def _regex_flags(self):
        return re.IGNORECASE

    def _act_name(self):
        return "Abono de Permanência"

    def _load_model(self):
        f_path = os.path.dirname(__file__)
        f_path += '/models/abono_ner.pkl'
        return joblib.load(f_path)

    def _props_names(self):
        return ["Tipo do Ato", "Nome do Servidor", "Matrícula",
                "Cargo Efetivo", "Classe", "Padrão",
                "Quadro pessoal permanente ou Suplementar",
                "Fundamento Legal do abono de permanência", "Órgão",
                "Processo GDF/SEI", "Vigencia", "Matricula SIAPE"]

    def _rule_for_inst(self):
        start = r"(Abono\sDE\sPERMANENCIA\s[(ao|equiva)][\s\S]*?)\s"
        body = r"([\s\S]*?"
        end = r"\.\n)"
        return start + body + end

    def _prop_rules(self):
        siape = r"[S|s][I|i][A|a][P|p][E|e]\s[N|n]?[o|O]?\s([\s\S]*?)[,| | .]"
        rules = {"nome": r"\s([^,]*?),\smatricula",
                 "matricula": r"matricula\s?n?o?\s([\s\S]*?)[,|\s]",
                 "cargo": r"Cargo\s[d|D]?[e|E]?\s([\s\S]*?),",
                 "classe": r"[C|c]lasse\s([\s\S]*?),",
                 "padrao": r"[p|P]adr[a|ã]o\s([\s\S]*?),",
                 "quadro": r"d?[e|a|o]?(Quadro[\s\S]*?)[,|;|.]",
                 "fundamento": r"nos\stermos\sdo\s([\s\S]*?),\sa?\s",
                 "orgao": r"Lotacao: ([\s\S]*?)[.]",
                 "sei": r"Processo SEI: ([\s\S]*?)\.\n",
                 "vigencia": r"a contar de ([\s\S]*?)\,",
                 "siape": siape}
        return rules



"""Functions to extract boxes from text."""

import fitz


def _extract_page_lines_content(page):
    """Extracts page lines.

    Args:
        page: fitz.fitz.Page object to have its bold content extracted.

    Returns:
        List[tuple(float, float, float, float, str)]
        A list containing lines content at the page, along with
        its bounding boxes.

    """
    lis = []
    for bl in page.getTextPage().extractDICT()['blocks']:
        for line in bl['lines']:
            for span in line['spans']:
                del span['color']
                del span['flags']
                lis.append((*span['bbox'], span['text']))
    return lis


def get_doc_text_boxes(doc: fitz.Document):
    """Returns list of list of extracted text blocks.

    Args:
        doc: an opened fitz document.

    Returns:
        List[List[tuple(float, float, float, float, str, int, int)]]

    """

    return [page.getTextBlocks() for page in doc]


def get_doc_text_lines(doc: fitz.Document):
    """Returns list of list of extracted text lines.

    Args:
        doc: an opened fitz document.

    Returns:
        List[List[tuple(float, float, float, str)]]

    """

    return [_extract_page_lines_content(page) for page in doc]


def _get_doc_img(doc: fitz.Document):
    """Returns list of list of image items.

    Note:
        This function is not intented to be used by final users,
        but internally. Image `items` are described at:

        https://pymupdf.readthedocs.io/en/latest/page/#Page.getImageBbox

    Args:
        doc: an opened fitz document

    Returns:
        List[List[tuple(int, int, int, int, str, str, str, str, int)]]
        (xref, smask, width, height, bpc, colorspace,
         alt. colorspace, filter, invoker)
    """

    return [page.getImageList(full=True) for page in doc]


def get_doc_img_boxes(doc: fitz.Document):
    """Returns list of list of bouding boxes of extracted images.

    Args:
        doc: an opened fitz document

    Returns:
        List[List[Rect(float, float, float, float)]]. Each Rect represents
            an image bounding box.

    """

    img_ll = _get_doc_img(doc)
    for idx, img_lis in enumerate(img_ll):
        page = doc[idx]
        img_ll[idx] = [page.getImageBbox(img) for img in img_lis]

    return img_ll



"""Regex backend for act and propriety extraction.

This module contains the ActRegex class, which have all that is necessary to
extract an act and, its proprieties, using regex rules.

"""

import re


class ActRegex:
    """Act Regex Class.

    This class encapsulate all functions, and attributes related
    to the process of regex extraction.

    Note:
        This class is one of the fathers of the Base act class.

    Attributes:
        _flags: All the regex flags which will be used in extraction.
        _rules: The regex rules for proprieties extraction.
        _inst_rule: The regex rule for act extraction.

    """

    def __init__(self):
        super(ActRegex, self).__init__()
        self._flags = self._regex_flags()
        self._rules = self._prop_rules()
        self._inst_rule = self._rule_for_inst()

    def _rule_for_inst(self):
        """Rule for extraction of the act

        Warning:
            Must return a regex rule that finds an act in two parts,
            containing a head and a body. Where only the body will be used
            to search for proprieties.

        Raises:
            NotImplementedError: Child class needs to overwrite this method.

        """
        raise NotImplementedError

    def _prop_rules(self):
        """Rules for extraction of the proprieties.

        Must return a dictionary of regex rules, where the key is
        the propriety type and the value is the rule.

        Raises:
            NotImplementedError: Child class needs to overwrite this method

        """
        raise NotImplementedError

    def _regex_flags(self):
        """Flag of the regex search"""
        return 0

    def _regex_instances(self):
        """Search for all instances of the act using the defined rule.

        Returns:
            List of all act instances in the text.
        """
        found = re.findall(self._inst_rule, self._text, flags=self._flags)
        results = []
        for instance in found:
            head, body = instance
            self.acts_str.append(head+body)
            results.append(body)

        return results

    def _find_prop_value(self, rule, act):
        """Find a single proprietie in an single act.

        Args:
            rule (str): The regex rule to search for.
            act (str): The act to apply the rule.

        Returns:
            The found propriety, or a nan in case nothing is found.

        """
        match = re.search(rule, act, flags=self._flags)
        if match:
            return tuple(x for x in match.groups() if x is not None)
        return "nan"

    def _regex_props(self, act_raw):
        """Create an act dict with all its proprieties.

        Args:
            act_raw (str): The raw text of a single act.

        Returns:
            The act, and its props in a dictionary format.

        """
        act = {}
        act["tipo_ato"] = self._name
        for key in self._rules:
            try:
                act[key], = self._find_prop_value(self._rules[key], act_raw)
            except Exception:
                act[key] = "nan"

        return act

    def _extract_instances(self):
        """Extract instances of an act.

        Warning:
            Instance must have an head and an body.

        Returns:
            All the instances of the act found.

        """
        found = self._find_instances()
        results = []
        for instance in found:
            head, body = instance
            self.acts_str.append(head+body)
            results.append(body)

        return results
